use Statistics::R;
# sequence
# perl /www/ePOSSUM/cgi-bin/PredictTFBSquot.pl mt=ATTTTTTTTTTTTTTCTTG tfs=all coordinates= wt=ATGCCTGTAAGCTGTCCT html_output=ePOSSUM_temp_1447169813_4186.html
# position
#  perl PredictTFBSquot.pl tfs=E2F1 coordinates=20:1111000-1111200 wt= html_output=ePOSSUM_temp_1447865327_14057.html
# VCF:
#  perl /www/ePOSSUM/cgi-bin/PredictTFBSquot.pl mt= tfs=known wt= chr=1 new=C old=CTG start=713950 html_output=ePOSSUM_temp.html

# select transcription_factor,tfbs_start,tfbs_end from encode.tfbs tfbs,encode.transcription_factors tf where tf.tf_number=tfbs.tf_number AND tfbs_chromosome=1 AND tfbs_start<=713935 AND tfbs_end>=713970


# select description,feature_type from ensembl_regulation75.feature_types where description like '%Factor Binding' GROUP BY description,feature_type ORDER BY description



use strict;
use ePOSSUM;

my $dbh='';
my @types = ( 'wt', 'mt' );
#my %strand_translation= (
#	minus => '-',
#	plus => '+)-strand',
#	both => 'both strands',
#	);
my %case_translation= (
	mt => 'variant',
	wt => 'wild-type',
	);	
my $output = new ePOSSUM;
my %FORM;
foreach (@ARGV) {
	my ( $param, $value ) = split /=/;
	$FORM{$param} = $value;
}
my $known_tfs = [
	qw /BCL11A BRCA1 CEBPB JUN CEBPD CREB1 CTCF E2F1 E2F4 E2F6 EBF1 EGR1 ELF1 ELK1 ELK4 ESR1 ETS1 FOS FOSL1 FOSL2 FOXA1 FOXM1 FOXP2 GABPA GATA2 GATA3 GTF2B GTF3C2 HDAC8 HNF4A HNF4G HSF1 IRF1 IRF3 IRF4 JUN JUNB JUND MAFF MAFK MAX MEF2A MEF2C MYBL2 NFATC1 NFE2 NFIC NFYA NFYB NR2C2 NR3C1 NRF1 PAX5 POU2F2 PRDM1 RBBP5 RELA REST RFX5 RUNX3 RXRA SP1 SP2 SP4 SPI1 SRF STAT1 STAT3 TCF3 TCF7L2 TEAD4 TFAP2A TFAP2C THAP1 USF1 USF2 YY1 ZBTB33 ZBTB7A ZEB1 ZNF143 ZNF263/ # /
];
my @tfs = ();

if ( $FORM{tfs} eq 'all' ) {
#	print "all\n";
	@tfs = @$known_tfs;
}
elsif ( $FORM{tfs} eq 'known' ) {
#	print "all\n";
	@tfs = 'known';
}
else {
	foreach my $tf ( split /,/, $FORM{tfs} ) {
		unless ( $tf ~~ $known_tfs ) {
			die2 ("No model for $tf");
		}
		else {
			push @tfs, $tf;
		#	print "tf\n";
		}
	}
}


my (%pv)=();
my $outputfile = '/data/tmp/' . $FORM{html_output};    #/
my $table_id=$FORM{html_output};
$table_id=~s/\.html$/\.txt/;
my $tablefile='/data/tmp/' .$table_id;
my $table_link='/temp/'.$table_id;

open (IN,'<','Bayes_PPV_NPV_sum.tsv') || die2 ($!);
$_=(<IN>);
while (<IN>) {
	#source	tf	matrix	prob	tp	tn	fp	fn	ppv	npv
	chomp;
	my ($source,$tf,$matrix,$prob,$tp,$tn,$fp,$fn,$ppv,$npv)=split /\t/;
	$prob=100-$prob;
	next unless $tf;
	if ($prob>50) {
		$pv{$tf}->{$source}->[$matrix]->[$prob]=$npv;
	}
	if ($prob<=50) {
		$pv{$tf}->{$source}->[$matrix]->[$prob]=$ppv;
	}	
}
close IN;

$output->StartOutput(
	"TFBS analysis running...",
	{
		refresh  => [ '/temp/' . $FORM{html_output}, 4 ],
		filename => $outputfile,
		styles => \"table {border:solid 1px grey}
		.bor {border: solid 1px grey;}
		.bc {border-collapse: collapse; }
		"
	}
);


my %x=();
my %searchable_tf=(); 
while (<DATA>) {
	next unless $_;
	if (/^(\d+);"(.*?)"/) {
		$searchable_tf{$2}=$1;
	} else {
		my ($tf)=$1 if /^"(.*?)"/;
		$x{$tf}.=$_;
	}
}   
CheckTFintersection() if $FORM{coordinates};
my %tfbs=();
if ($FORM{chr}) {
	print "Retrieving sequence from database...<br>\n";
	my $dbh=$output->Connect();
	my ($chr,$start,$end,$old,$new,$length)=(@FORM{qw /chr start end old new/},0); #/
	if ($chr and $start and $end){ 
		$length=$end-$start+1;
	}
	elsif ($chr and $start and $old and $new) {
		$start-=15;
		$length=30+(length $old);
		$end=$start+$length+1;
	}
	else {
		die2 ("Wrong format!");
	}

	die2("Format error","Coordinates '$FORM{coordinates}' cannot be handled.") unless $length>0 ;
	my $q_seq_sql="SELECT substr(sequence,?,?) FROM genome.sequence WHERE chromosome=? ";
	my $q_seq=$dbh->prepare($q_seq_sql) || die2("DB error",$DBI::errstr);
	$q_seq->execute($start,$length,$chr)  || die2("DB error",$DBI::errstr);
	my $seq=$q_seq->fetchrow_arrayref  || die2("DB error",$DBI::errstr);
	die2("error","Position '$FORM{coordinates}' is not covered in genome version 37") unless @$seq and length $seq->[0];
	$FORM{wt}=$seq->[0];
	if ($old and $new) {
		my $ref=substr($seq->[0],15,(length $old));
		$FORM{wt}=substr($seq->[0],0,15).$old.substr($seq->[0],(15+length $old));
		$FORM{mt}=substr($seq->[0],0,15).$new.substr($seq->[0],(15+length $old));
		die2("Genome version error","Your reference bases '$old' do not match the genomic sequence '$ref' - ePOSSUM uses build 37!") unless $old eq $ref; 
	}
	$FORM{location}="$chr: $start-$end";
	print "Retrieving known TFBS from database ($chr: $start-$end)...<br>\n";
	my $q_tf_sql="SELECT transcription_factor, tfbs_start, tfbs_end FROM encode.tfbs pos, encode.transcription_factors tf
	WHERE tf.tf_number=pos.tf_number AND tfbs_chromosome=? AND tfbs_end>=? AND tfbs_start<=?";
	my $q_tf=$dbh->prepare($q_tf_sql) || die2("DB error",$DBI::errstr);
	$q_tf->execute($chr,$start,$end)  || die2("DB error",$DBI::errstr);
	my $known_tfbs=$q_tf->fetchall_arrayref  || die2("DB error",$DBI::errstr);
	foreach my $tfbsref (@$known_tfbs) {
		my ($tfbs,$start,$end)=@$tfbsref;
		push @{$tfbs{$tfbs}},"$start-$end" if $tfbs~~$known_tfs;
	}
	print scalar keys %tfbs," known TFBSs possibly affected.<br>\n";
	if ($tfs[0] eq 'known') {
		@tfs=keys %tfbs;
		print "Restricting screening to known TFBSs possibly affected by your variant...<br>\n";
		die2 ("NONE found!") unless @tfs;
		
	}

	
}
# GGCCCTCACACTCTCCAACCTCATCTCCCTGGTGAGAGGCC
#      TCACACTCTCCAACCTCATCTCCCTGGTGAG

die2("Illegal bases in wild-type sequence: $FORM{wt}")
  if $FORM{wt} and $FORM{wt} !~ /^[ACGT]+$/;
die2("Illegal bases in variant sequence: $FORM{mt}")
  if $FORM{mt} and $FORM{mt} !~ /^[ACGT]+$/;
die2("No sequence(s).") unless $FORM{wt};



print "Searching binding motifs for ", join( ", ", @tfs ), "<br>\n";

print qq !Wild-type sequence: $FORM{wt}<br>\n!;
print qq !Variant sequence: $FORM{mt}<br>\n!;
print qq !Starting R.<br>This will take many seconds - please be as patient as possible (you are free to hit F5/reload)...<br>\n!; #/

my $starttime = time();


sub CheckTFintersection {
	# nur um zu pruefen, ob die DB-TFBS mit den Matrizen uebereinstimmen
	return; 
	foreach my $tf (@$known_tfs) {
		print "<b>$tf</b><br>\n" unless  $searchable_tf{$tf};
		print "$tf: $searchable_tf{$tf}<br>\n";
	}
	exit 0;
}

#my $R =();
my $R = InvokeR(); # Create a communication bridge with R and start R

my %results = ();
my @sources = ( "hPDI", "JASPAR", "HT-SELEX" );
my %translation =
 ( "hPDI"=>'hPDI',
  "JASPAR" =>'JASPAR_2014',
   "HT-SELEX"=>'Jolma2013' );
my %translation2 =
 ( "hPDI"=>'hPDI',
  "JASPAR_2014" =>'JASPAR',
   "Jolma2013"=>'HT-SELEX' );

foreach my $tf (@tfs) {    # /
	my $starttime2 = time();
	my $x='';
	if ($R) {
		$R->set( 'tf', $tf );
		$R->run_from_file('PredictQuot.R');
		$x = $R->run(
			'write.table(bayes_result, append=F, sep=",", row.names=F, col.names=T)  '
		);
		print "processing $tf...<br>\n";
	}
	else {
		$x=$x{$tf};
	}
	$x =~ s/"//g;
	foreach ( split /\n/, $x ) {
		next if /^TF_b/;
		# pred: Wildtyp im Zaehler -> >1 bei gain 
# TF_b,mat_b,db_b,maxscore_wt, maxscore_mt, startpos_wt_b, endpos_wt_b, startpos_mt_b, endpos_mt_b, maxstrand_b, maxstrand_b_mt, pred, bayesscore_gain,wtzquot_b)
		
		my ($thistf,    $matrix_no,   $source,    $wt_score,	$mt_score,  $startpos_wt,
			$endpos_wt, $startpos_mt,	$endpos_mt, $strand_wt,   $strand_mt, $prediction,	$pred_p
			) = split /,/;
		$matrix_no--;
		$source = $translation2{$source};
		if ($wt_score==$mt_score){
			$pred_p='';
			$prediction='no change';
		}
		$results{$tf}->{$source}->[$matrix_no] = [ $wt_score, $mt_score, $prediction,	$pred_p, $startpos_wt,
			$endpos_wt, $startpos_mt,	$endpos_mt, $strand_wt,   $strand_mt ];
	}
	my $endtime = time();
	print "<small>Done (", $endtime - $starttime2, " seconds - ", $endtime - $starttime, " seconds from start).<br></small>\n";
}
$R->stop() if $R;
close STDOUT;


$output->StartOutput("TFBS analysis results", {
		filename => $outputfile,
		styles => \"table {border:solid 1px grey}
		.bor {border: solid 1px grey;}
		.bc {border-collapse: collapse; }
		.greyl {color: #D0D0D0}
		"
	});

print qq ! <div id="doku" style="border: 0px none; overflow: visible; position: absolute; visibility: visible; z-index: 99; 
display: show; left: 0px; top: 0px; text-align:right; width:100% ; "><A HREF="http://mutationtaster.charite.de/ePOSSUM/ePOSSUM_results_info.html">documentation</A> </div>
<div id="popup" style="border: 1px solid black; overflow: visible; position: absolute;  
visibility: visible; z-index: 99; display: none; left: -28px; top: 708px;  border-radius: 8px; padding: 4px; "
	onMouseDown="firstcorner(this)"> </div>
<script src="/javascript/ePOSSUM.js" type="text/javascript"></script> \n!; #"
print qq !<h1>$FORM{name}</h1>\n! if $FORM{name};
print qq !<hr>
<table border="0" width="100%" style="border: 0px solid white">
	<tr>
		<td>
			<table cellpadding="3" cellspacing="0" style="border-collapse: collapse;  border: 1px solid black;">
				<tr><td>Wild-type sequence</td><td>$FORM{wt}</td></tr>
				<tr><td>Variant sequence</td><td>$FORM{mt}</td></tr>!; 
				if (	$FORM{location}){
					print qq !<tr><td>Location</td><td>$FORM{location}</td></tr>\n!;
				}				
print qq !</table></td><td><td><A HREF="$table_link">output as table</A></td></tr>
</table><br>
<table cellpadding="3" cellspacing="0" style="border: 0px solid black;"><tr style="border-bottom: 1px solid black;">
<td><h2>TF</h2></td> <td colspan="9"><h2>predictions</h2></td></tr>\n!;
my $pssms=0;

open (TABLE,'>',$tablefile) || die2 ($!);
print TABLE join ("\t",qw !tf source wt_score mut_score prediction bayes_probability_gain ppv npv startpos_wt endpos_wt strand_wt	
startpos_mut endpos_mut strand_mut result!),"\n";

foreach my $tf ( sort keys %results ) {
	my $found_this_one=($tfbs{$tf}?'known TFBS':'');
	print "<tr><td><h2>$tf</h2>$found_this_one</td>";
	my %comb_pred = ();
	my %effects=();
	my @table_output=();
 	SOURCES: foreach my $source (@sources) {
		next SOURCES unless $results{$tf}->{$source}; 		
		my $results = scalar @{ $results{$tf}->{$source} };
		foreach my $matrix_no ( 0 .. ( $results - 1 ) ) {
			my $mat2=($results>1?' matrix #'.($matrix_no+1):'');
			$pssms++;
			# $mat2 waere die Matrix-Nr
			my %this_score = ();
			my (%endpos,%startpos,%strand)=();
			my (@pv)=();
			@{this_score}{@types} =  @{ $results{$tf}->{$source}->[$matrix_no] }[ 0, 1 ];    # /
			my ($pred,$bayes_p)=  @{ $results{$tf}->{$source}->[$matrix_no] }[ 2, 3 ];    # /
			($startpos{wt},$endpos{wt}, $startpos{mt},	$endpos{mt}, $strand{wt},   $strand{mt}) =  @{$results{$tf}->{$source}->[$matrix_no]}[4..9];    # /
			my $text='';
			my $colour='';
			my $fg_colour='';
			my $bayes_o=sprintf ("%1.3f",$bayes_p);
			my @table_ppv_npv=('n/a','n/a');
			unless ($this_score{wt} or $this_score{mt}) {
				$pred='no binding';
				$text='no binding';
			}
			else {
				$text=	"<b>best hits for $tf with model from $source $mat2</b><br><table border=0 class=bc><tr><td class=bor>sequence</td><td class=bor>highest score</td><td class=bor>strand</td><td class=bor>start</td><td class=bor>end</td></tr>";
				my $prob=sprintf("%2d", (1-$bayes_p)*100);
				$pv[1]=sprintf ("%1.3f",($pv{$tf}->{$source}->[$matrix_no]->[$prob]*100) );
				my $saturation=int(155-$pv[1]*55/50);
				$saturation=100 if $saturation>100;
				push @{$effects{$pred}},$pv[1] if $pv[1] >= 70;
				die2 ("$tf : $source : $bayes_o : $pv[1] : $saturation ")if $saturation>100 or $saturation<0;
				if ($pred eq 'loss') { 
					$bayes_p=1-$bayes_p;
					$pv[0]='npv';
					$table_ppv_npv[1]=$pv[1];
					$colour='background-color:hsl(12, 100%, '.$saturation.'%);';
				}
				elsif ($pred eq 'gain') { 
					$pv[0]='ppv';
					$table_ppv_npv[0]=$pv[1];
					$fg_colour='class="greyl"' if $saturation<75;
					$colour='background-color:hsl(224, 100%, '.$saturation.'%);';
				}
				else  { 
					$pv[1]='n/a';
				}				
				foreach my $type (@types) {
					$this_score{$type}=sprintf ("%1.3f",$this_score{$type});
					$text.=	'<tr><td class=bor>'.join ("</td><td class=bor>",$case_translation{$type},$this_score{$type},$strand{$type},$startpos{$type},$endpos{$type})."</td></tr>";
				}
				$text.= '</table><small><i>positions refer to (+)-strand</small></i>';
			}
			my $bwidth=($found_this_one ?'3px solid black':'1px solid grey');
	
			print qq !<td style="vertical-align: middle">
			<table  style="border: $bwidth ; border-radius: 8px;width:8em; height:6em; $colour" $fg_colour
			onMouseOver="p('$text')"
			onMouseOut="k('')">
			<tr style="border-bottom: 2px solid grey;"><td colspan="2" style="border-bottom: 2px solid grey; text-align: right"><i>$source</i></td></tr>
			
			<tr><td colspan="2" style="text-align: center"><b>$pred</b></td></tr>
			<tr><td>p</td><td style="text-align: right">!,(sprintf "%2.1f%%</td></tr>\n",$bayes_p*100);
			push @table_output, join ("\t",$tf,$source,$this_score{wt}, $this_score{mt}, $pred,	$bayes_o,  @table_ppv_npv,
			$startpos{wt},$endpos{wt}, $strand{wt},  $startpos{mt},	$endpos{mt}, $strand{mt});
			if (@pv) {
				print qq !<tr><td>$pv[0]</td><td style="text-align: right">!,(sprintf "%2.1f%%</td></tr>\n",$pv[1]);	
			}
			foreach my $type (@types) { 
				print qq !<tr><td>$type</td><td style="text-align: right">!.$this_score{$type}."</td></tr>\n";
			}
			print "</table></td>";
		}
 	}
	my $result='';
 	if ($tfbs{$tf}) {
 		print qq !<td style="vertical-align: middle">
			<table  style="border:1px solid black ; border-radius: 8px; height:6em; width:8em; background-color:#CCCCCC">\n!;

		if ($effects{loss} and ! $effects{gain}){
			$result='<b class="red">attenuation / loss of known TFBS</b>';
		}
		elsif ($effects{loss} and $effects{gain}){
			$result='contradicting predictions for known TFBS';
		}
		elsif (! $effects{loss} and $effects{gain}){
			$result='<b class="blue">enhancement of known TFBS</b>';
		}
		else {
			$result='no definite results';
		}
		print qq !<tr><td>$result</td></tr></table></td>\n!;
 	}		
 	else {

 		print qq !<td style="vertical-align: middle">
			<table  style="border:1px solid black ; border-radius: 8px; height:6em; width:8em; background-color:white">\n!;
		if ($effects{loss} and ! $effects{gain}){
			$result='<b class="red">attenuation / loss of unconfirmed TFBS</b>';
		}
		elsif ($effects{loss} and $effects{gain}){
			$result='contradicting predictions for this site';
		}
		elsif (! $effects{loss} and $effects{gain}){
			$result='<b class="blue">creation of new TFBS</b>';
		}
		else {
			$result='no definite results';
		}
		print qq !<tr><td>$result</td></tr></table></td>\n!;
 	}		 
 	foreach my $outputline (@table_output) {
 		$result=~s/<.*?>//g;
 		print TABLE $outputline,"\t",$result,"\n";
 	}
	print "</tr>";						
}


close TABLE;

my $info=(scalar keys %results)." TFs - $pssms PSSMs.";
print "</table><hr>$info</body></html>\n";
exit 0;

sub InvokeR {
	my $R = Statistics::R->new( shared => 0 );
	$R->run(qq !library(Biostrings) !);
	$R->run(qq !library(MotifDb) !);
	$R->run(qq !library("klaR") !);
	$R->run(qq ! rm(list = ls()) !);
	$R->set( 'wt', $FORM{wt} );
	$R->set( 'mt', $FORM{mt} );
	return $R;	
}



sub die2 {
	my $diefile=$outputfile;
	$diefile=~s/\.html/.fail/;
	open (DIE,'>',$diefile) || die ($!);
	print DIE join ("\n",@_);
	close DIE;
	$output->PegOut(@_);
}

__DATA__
131;"ARID3A"
56;"ATF1"
29;"ATF2"
44;"ATF3"
87;"BACH1"
92;"BATF"
134;"BCL11A"
82;"BCL3"
72;"BCLAF1"
152;"BDP1"
16;"BHLHE40"
45;"BRCA1"
161;"BRF1"
147;"BRF2"
71;"CBX3"
86;"CCNT2"
2;"CEBPB"
132;"CEBPD"
114;"CHD1"
42;"CHD2"
55;"CREB1"
111;"CTBP2"
3;"CTCF"
64;"CTCFL"
27;"E2F1"
52;"E2F4"
19;"E2F6"
63;"EBF1"
8;"EGR1"
54;"ELF1"
39;"ELK1"
119;"ELK4"
37;"EP300"
104;"ESR1"
126;"ESRRA"
96;"ETS1"
98;"EZH2"
157;"FAM48A"
59;"FOS"
77;"FOSL1"
17;"FOSL2"
9;"FOXA1"
123;"FOXA2"
136;"FOXM1"
43;"FOXP2"
5;"GABPA"
75;"GATA1"
78;"GATA2"
80;"GATA3"
49;"GRp20"
143;"GTF2B"
28;"GTF2F1"
109;"GTF3C2"
118;"HDAC1"
88;"HDAC2"
144;"HDAC6"
139;"HDAC8"
94;"HMGN3"
122;"HNF4A"
107;"HNF4G"
138;"HSF1"
148;"IKZF1"
36;"IRF1"
141;"IRF3"
133;"IRF4"
51;"JUN"
99;"JUNB"
18;"JUND"
115;"KAP1"
31;"KDM5A"
34;"KDM5B"
14;"MAFF"
15;"MAFK"
20;"MAX"
11;"MAZ"
155;"MBD4"
84;"MEF2A"
85;"MEF2C"
135;"MTA3"
32;"MXI1"
128;"MYBL2"
30;"MYC"
116;"NANOG"
83;"NFATC1"
140;"NFE2"
129;"NFIC"
46;"NFYA"
48;"NFYB"
73;"NR2C2"
103;"NR2F2"
112;"NR3C1"
120;"NRF1"
22;"PAX5"
69;"PBX3"
23;"PHF8"
24;"PML"
21;"POLR2A"
160;"POLR3G"
33;"POU2F2"
149;"POU5F1"
151;"PPARGC1A"
156;"PRDM1"
12;"RAD21"
74;"RBBP5"
76;"RCOR1"
158;"RDBP"
47;"RELA"
50;"REST"
40;"RFX5"
150;"RPC155"
10;"RUNX3"
108;"RXRA"
113;"SAP30"
95;"SETDB1"
105;"SIN3A"
26;"SIN3AK20"
153;"SIRT6"
57;"SIX5"
145;"SMARCA4"
125;"SMARCB1"
124;"SMARCC1"
154;"SMARCC2"
13;"SMC3"
7;"SP1"
62;"SP2"
100;"SP4"
93;"SPI1"
130;"SREBP1"
53;"SRF"
91;"STAT1"
127;"STAT2"
121;"STAT3"
117;"STAT5A"
110;"SUZ12"
4;"TAF1"
38;"TAF7"
79;"TAL1"
60;"TBL1XR1"
35;"TBP"
81;"TCF12"
89;"TCF3"
41;"TCF7L2"
65;"TEAD4"
101;"TFAP2A"
102;"TFAP2C"
66;"THAP1"
106;"TRIM28"
70;"UBTF"
6;"USF1"
58;"USF2"
142;"WRNIP1"
25;"YY1"
1;"ZBTB33"
97;"ZBTB7A"
67;"ZEB1"
146;"ZKSCAN1"
61;"ZNF143"
137;"ZNF217"
68;"ZNF263"
90;"ZNF274"
159;"ZZZ3"
